# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from .autompg import autompg_data
from .boston_housing import boston_housing_data
from .iris import iris_data
from .local_mnist import loadlocal_mnist
from .mnist import mnist_data
from .multiplexer import make_multiplexer_dataset
from .three_blobs import three_blobs_data
from .wine import wine_data

__all__ = [
    "iris_data",
    "wine_data",
    "autompg_data",
    "loadlocal_mnist",
    "mnist_data",
    "boston_housing_data",
    "three_blobs_data",
    "make_multiplexer_dataset",
]



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from mlxtend._base import _BaseModel, _Regressor
from mlxtend.utils import assert_raises


class BlankRegressor(_BaseModel, _Regressor):
    def __init__(self):
        pass

    def _fit(self, X, y, init_params=True):
        pass

    def _predict(self, X):
        pass


def test_float_ok():
    y = np.array([1.0, 2.0])
    reg = BlankRegressor()
    reg._check_target_array(y=y)


def test_float_fail():
    y = np.array([1, 2], dtype=np.int_)
    reg = BlankRegressor()
    assert_raises(
        AttributeError,
        f"y must be a float array.\nFound {str(y.dtype)}",
        reg._check_target_array,
        y,
    )


def test_predict_fail():
    X = np.array([[1], [2], [3]])
    est = BlankRegressor()
    est._is_fitted = False
    assert_raises(AttributeError, "Model is not fitted, yet.", est.predict, X)


def test_predict_pass():
    X = np.array([[1], [2], [3]])
    y = np.array([1.0, 2.0, 3.0])
    est = BlankRegressor()
    est.fit(X, y)
    est.predict(X)


def test_fit_1():
    X = np.array([[1], [2], [3]])
    est = BlankRegressor()
    assert_raises(
        TypeError, "fit() missing 1 required positional argument: 'y'", est.fit, X
    )


def test_fit_2():
    X = np.array([[1], [2], [3]])
    y = np.array([1.0, 2.0, 3.0])
    est = BlankRegressor()
    est.fit(X=X, y=y)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import pandas as pd
from numpy.testing import assert_almost_equal
from packaging.version import Version
from sklearn import __version__ as sklearn_version
from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import GroupKFold
from sklearn.neighbors import KNeighborsClassifier

from mlxtend.classifier import SoftmaxRegression
from mlxtend.data import boston_housing_data
from mlxtend.feature_selection import ExhaustiveFeatureSelector as EFS
from mlxtend.utils import assert_raises


def dict_compare_utility(d1, d2, decimal=2):
    assert d1.keys() == d2.keys(), "%s != %s" % (d1, d2)
    for i in d1:
        err_msg1 = "d1[%s]['feature_idx']" " != d2[%s]['feature_idx']" % (i, i)
        err_msg2 = "d1[%s]['feature_names']" " != d2[%s]['feature_names']" % (i, i)
        assert d1[i]["feature_idx"] == d2[i]["feature_idx"], err_msg1
        assert d1[i]["feature_names"] == d2[i]["feature_names"], err_msg2
        assert_almost_equal(
            d1[i]["avg_score"],
            d2[i]["avg_score"],
            decimal=decimal,
            err_msg=("d1[%s]['avg_score']" " != d2[%s]['avg_score']" % (i, i)),
        )
        assert_almost_equal(
            d1[i]["cv_scores"],
            d2[i]["cv_scores"],
            decimal=decimal,
            err_msg=("d1[%s]['cv_scores']" " != d2[%s]['cv_scores']" % (i, i)),
        )


def test_minfeatures_1():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier()

    efs = EFS(estimator=knn, min_features=0, max_features=2)
    expect = "min_features must be smaller than 5 and larger than 0"
    assert_raises(AttributeError, expect, efs.fit, X, y)


def test_maxfeatures_1():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier()

    efs = EFS(estimator=knn, min_features=1, max_features=0)
    expect = "max_features must be smaller than 5 and larger than 0"
    assert_raises(AttributeError, expect, efs.fit, X, y)


def test_minmaxfeatures_1():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier()

    efs = EFS(estimator=knn, min_features=3, max_features=2)
    expect = "min_features must be <= max_features"
    assert_raises(AttributeError, expect, efs.fit, X, y)


def test_knn_wo_cv():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = EFS(
        knn,
        min_features=2,
        max_features=3,
        scoring="accuracy",
        cv=0,
        print_progress=False,
    )
    efs1 = efs1.fit(X, y)
    expect = {
        0: {
            "feature_idx": (0, 1),
            "feature_names": ("0", "1"),
            "avg_score": 0.8333333333333334,
            "cv_scores": np.array([0.8333333333333334]),
        },
        1: {
            "feature_idx": (0, 2),
            "feature_names": ("0", "2"),
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.95999999999999996]),
        },
        2: {
            "feature_idx": (0, 3),
            "feature_names": ("0", "3"),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
        3: {
            "feature_idx": (1, 2),
            "feature_names": ("1", "2"),
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.96]),
        },
        4: {
            "feature_idx": (1, 3),
            "feature_names": ("1", "3"),
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.96]),
        },
        5: {
            "feature_idx": (2, 3),
            "feature_names": ("2", "3"),
            "avg_score": 0.97333333333333338,
            "cv_scores": np.array([0.97333333]),
        },
        6: {
            "feature_idx": (0, 1, 2),
            "feature_names": ("0", "1", "2"),
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.96]),
        },
        7: {
            "feature_idx": (0, 1, 3),
            "feature_names": ("0", "1", "3"),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
        8: {
            "feature_idx": (0, 2, 3),
            "feature_names": ("0", "2", "3"),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
        9: {
            "feature_idx": (1, 2, 3),
            "feature_names": ("1", "2", "3"),
            "avg_score": 0.97333333333333338,
            "cv_scores": np.array([0.97333333]),
        },
    }
    dict_compare_utility(d1=expect, d2=efs1.subsets_)


def test_knn_cv3():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = EFS(
        knn,
        min_features=3,
        max_features=3,
        scoring="accuracy",
        cv=4,
        print_progress=False,
    )
    efs1 = efs1.fit(X, y)
    expect = {
        0: {
            "avg_score": 0.9329658605974395,
            "feature_idx": (0, 1, 2),
            "feature_names": ("0", "1", "2"),
            "cv_scores": np.array([0.974, 0.947, 0.892, 0.919]),
        },
        1: {
            "avg_score": 0.9400782361308677,
            "feature_idx": (0, 1, 3),
            "feature_names": ("0", "1", "3"),
            "cv_scores": np.array([0.921, 0.947, 0.919, 0.973]),
        },
        2: {
            "avg_score": 0.9532361308677098,
            "feature_idx": (0, 2, 3),
            "feature_names": ("0", "2", "3"),
            "cv_scores": np.array([0.974, 0.947, 0.919, 0.973]),
        },
        3: {
            "avg_score": 0.97275641025641035,
            "feature_idx": (1, 2, 3),
            "feature_names": ("1", "2", "3"),
            "cv_scores": np.array([0.974, 1.0, 0.946, 0.973]),
        },
    }

    if Version(sklearn_version) < Version("1.0"):
        expect[0]["avg_score"] = 0.9391025641025641
        expect[0]["cv_scores"] = np.array([0.974, 0.947, 0.892, 0.946])
        expect[2]["avg_score"] = 0.9529914529914529

    if Version(sklearn_version) < Version("0.22"):
        expect[0]["cv_scores"] = np.array(
            [0.97435897, 0.94871795, 0.88888889, 0.94444444]
        )
        expect[1]["cv_scores"] = np.array(
            [0.92307692, 0.94871795, 0.91666667, 0.97222222]
        )
        expect[2]["cv_scores"] = np.array(
            [0.97435897, 0.94871795, 0.91666667, 0.97222222]
        )
        expect[3]["cv_scores"] = np.array(
            [0.97435897, 0.94871795, 0.91666667, 0.97222222]
        )
        expect[1]["avg_score"] = 0.94017094017094016
        assert round(efs1.best_score_, 4) == 0.9728
    else:
        assert round(efs1.best_score_, 4) == 0.9732

    dict_compare_utility(d1=expect, d2=efs1.subsets_)
    assert efs1.best_idx_ == (1, 2, 3)
    assert efs1.best_feature_names_ == ("1", "2", "3")


def test_knn_cv3_groups():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = EFS(
        knn,
        min_features=3,
        max_features=3,
        scoring="accuracy",
        cv=GroupKFold(n_splits=3),
        print_progress=False,
    )
    np.random.seed(1630672634)
    groups = np.random.randint(0, 6, size=len(y))
    efs1 = efs1.fit(X, y, groups=groups)

    expect = {
        0: {
            "cv_scores": np.array([0.97916667, 0.93877551, 0.9245283]),
            "feature_idx": (0, 1, 2),
            "avg_score": 0.9474901595858469,
            "feature_names": ("0", "1", "2"),
        },
        1: {
            "cv_scores": np.array([1.0, 0.93877551, 0.9245283]),
            "feature_idx": (0, 1, 3),
            "avg_score": 0.9544346040302915,
            "feature_names": ("0", "1", "3"),
        },
        2: {
            "cv_scores": np.array([0.97916667, 0.95918367, 0.9245283]),
            "feature_idx": (0, 2, 3),
            "avg_score": 0.9542928806742822,
            "feature_names": ("0", "2", "3"),
        },
        3: {
            "cv_scores": np.array([0.97916667, 0.95918367, 0.94339623]),
            "feature_idx": (1, 2, 3),
            "avg_score": 0.9605821888503829,
            "feature_names": ("1", "2", "3"),
        },
    }
    dict_compare_utility(d1=expect, d2=efs1.subsets_)


def test_fit_params():
    iris = load_iris()
    X = iris.data
    y = iris.target
    sample_weight = np.ones(X.shape[0])
    forest = RandomForestClassifier(n_estimators=100, random_state=123)
    efs1 = EFS(
        forest,
        min_features=3,
        max_features=3,
        scoring="accuracy",
        cv=4,
        print_progress=False,
    )
    efs1 = efs1.fit(X, y, sample_weight=sample_weight)
    expect = {
        0: {
            "feature_idx": (0, 1, 2),
            "feature_names": ("0", "1", "2"),
            "cv_scores": np.array([0.947, 0.868, 0.919, 0.973]),
            "avg_score": 0.9269203413940257,
        },
        1: {
            "feature_idx": (0, 1, 3),
            "feature_names": ("0", "1", "3"),
            "cv_scores": np.array([0.921, 0.921, 0.892, 1.0]),
            "avg_score": 0.9337606837606838,
        },
        2: {
            "feature_idx": (0, 2, 3),
            "feature_names": ("0", "2", "3"),
            "cv_scores": np.array([0.974, 0.947, 0.919, 0.973]),
            "avg_score": 0.9532361308677098,
        },
        3: {
            "feature_idx": (1, 2, 3),
            "feature_names": ("1", "2", "3"),
            "cv_scores": np.array([0.974, 0.947, 0.892, 1.0]),
            "avg_score": 0.9532361308677098,
        },
    }

    if Version(sklearn_version) < Version("0.22"):
        expect[0]["avg_score"] = 0.9401709401709402
        expect[0]["cv_scores"] = np.array(
            [0.94871795, 0.92307692, 0.91666667, 0.97222222]
        )
        expect[1]["cv_scores"] = np.array(
            [0.94871795, 0.92307692, 0.91666667, 0.97222222]
        )
        expect[2]["cv_scores"] = np.array(
            [0.94871795, 0.92307692, 0.91666667, 0.97222222]
        )
        expect[2]["avg_score"] = 0.9599358974358974
        expect[3]["avg_score"] = 0.9599358974358974
        expect[3]["cv_scores"] = np.array([0.97435897, 0.94871795, 0.91666667, 1.0])
        assert round(efs1.best_score_, 4) == 0.9599

    else:
        assert round(efs1.best_score_, 4) == 0.9532

    dict_compare_utility(d1=expect, d2=efs1.subsets_)
    assert efs1.best_idx_ == (0, 2, 3)


def test_regression():
    X, y = boston_housing_data()
    X = X[:, [1, 2, 6, 8, 12]]
    lr = LinearRegression()
    efs_r = EFS(
        lr,
        min_features=3,
        max_features=4,
        scoring="neg_mean_squared_error",
        cv=10,
        print_progress=False,
    )
    efs_r = efs_r.fit(X, y)
    assert efs_r.best_idx_ == (0, 2, 4)
    assert round(efs_r.best_score_, 4) == -40.8777


def test_clone_params_fail():
    class Perceptron(object):
        def __init__(self, eta=0.1, epochs=50, random_seed=None, print_progress=0):
            self.eta = eta
            self.epochs = epochs
            self.random_seed = random_seed
            self.print_progress = print_progress
            self._is_fitted = False

        def _fit(self, X, y, init_params=True):
            self._check_target_array(y, allowed={(0, 1)})
            y_data = np.where(y == 0, -1.0, 1.0)

            if init_params:
                self.b_, self.w_ = self._init_params(
                    weights_shape=(X.shape[1], 1),
                    bias_shape=(1,),
                    random_seed=self.random_seed,
                )
                self.cost_ = []

            rgen = np.random.RandomState(self.random_seed)
            for i in range(self.epochs):
                errors = 0

                for idx in self._yield_minibatches_idx(
                    rgen=rgen, n_batches=y_data.shape[0], data_ary=y_data, shuffle=True
                ):
                    update = self.eta * (y_data[idx] - self._to_classlabels(X[idx]))
                    self.w_ += (update * X[idx]).reshape(self.w_.shape)
                    self.b_ += update
                    errors += int(update != 0.0)

                if self.print_progress:
                    self._print_progress(
                        iteration=i + 1, n_iter=self.epochs, cost=errors
                    )
                self.cost_.append(errors)
            return self

        def _net_input(self, X):
            """Net input function"""
            return (np.dot(X, self.w_) + self.b_).flatten()

        def _to_classlabels(self, X):
            return np.where(self._net_input(X) < 0.0, -1.0, 1.0)

        def _predict(self, X):
            return np.where(self._net_input(X) < 0.0, 0, 1)

    expect = (
        "Cannot clone object. You should provide an"
        " instance of scikit-learn estimator instead of a class."
    )

    assert_raises(
        TypeError,
        expect,
        EFS,
        Perceptron,
        min_features=2,
        max_features=2,
        clone_estimator=True,
    )


def test_clone_params_pass():
    iris = load_iris()
    X = iris.data
    y = iris.target
    lr = SoftmaxRegression(random_seed=1)
    efs1 = EFS(
        lr,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        clone_estimator=False,
        print_progress=False,
        n_jobs=1,
    )
    efs1 = efs1.fit(X, y)
    assert efs1.best_idx_ == (1, 3)


def test_transform_not_fitted():
    iris = load_iris()
    X = iris.data
    knn = KNeighborsClassifier(n_neighbors=4)

    efs1 = EFS(
        knn,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        clone_estimator=False,
        print_progress=False,
        n_jobs=1,
    )

    expect = "ExhaustiveFeatureSelector has not been fitted, yet."

    assert_raises(AttributeError, expect, efs1.transform, X)


def test_fit_transform():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)

    efs1 = EFS(
        knn,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        clone_estimator=False,
        print_progress=False,
        n_jobs=1,
    )

    X_t = efs1.fit_transform(X, y)
    assert X_t.shape == (150, 2)


def test_get_metric_dict_not_fitted():
    knn = KNeighborsClassifier(n_neighbors=4)

    efs1 = EFS(
        knn,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        clone_estimator=False,
        print_progress=False,
        n_jobs=1,
    )

    expect = "ExhaustiveFeatureSelector has not been fitted, yet."

    assert_raises(AttributeError, expect, efs1.get_metric_dict)


def test_check_pandas_dataframe_fit():
    knn = KNeighborsClassifier(n_neighbors=4)
    iris = load_iris()
    X = iris.data
    y = iris.target
    efs1 = EFS(
        knn,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        clone_estimator=False,
        print_progress=False,
        n_jobs=1,
    )

    df = pd.DataFrame(
        X, columns=["sepal length", "sepal width", "petal length", "petal width"]
    )

    sfs1 = efs1.fit(X, y)
    assert efs1.best_idx_ == (2, 3), efs1.best_idx_
    assert efs1.best_feature_names_ == ("2", "3")
    assert efs1.interrupted_ is False

    sfs1._TESTING_INTERRUPT_MODE = True
    sfs1 = sfs1.fit(df, y)
    assert efs1.best_idx_ == (0, 1), efs1.best_idx_
    assert efs1.best_feature_names_ == ("sepal length", "sepal width")
    assert efs1.interrupted_ is True


def test_check_pandas_dataframe_transform():
    knn = KNeighborsClassifier(n_neighbors=4)
    iris = load_iris()
    X = iris.data
    y = iris.target
    efs1 = EFS(
        knn,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        clone_estimator=False,
        print_progress=False,
        n_jobs=1,
    )

    df = pd.DataFrame(
        X, columns=["sepal length", "sepal width", "petal length", "petal width"]
    )
    efs1 = efs1.fit(df, y)
    assert efs1.best_idx_ == (2, 3)
    assert (150, 2) == efs1.transform(df).shape


def test_knn_wo_cv_with_feature_groups_integer():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = EFS(
        knn,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        print_progress=False,
        feature_groups=[[0], [1, 2], [3]],
    )
    efs1 = efs1.fit(X, y)
    # expect is based on what provided in `test_knn_wo_cv`
    expect = {
        0: {
            "feature_idx": (0, 1, 2),
            "feature_names": ("0", "1", "2"),
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.96]),
        },
        1: {
            "feature_idx": (0, 3),
            "feature_names": ("0", "3"),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
        2: {
            "feature_idx": (1, 2, 3),
            "feature_names": ("1", "2", "3"),
            "avg_score": 0.97333333333333338,
            "cv_scores": np.array([0.97333333]),
        },
    }
    dict_compare_utility(d1=expect, d2=efs1.subsets_)


def test_knn_wo_cv_with_feature_groups_string():
    iris = load_iris()
    X = iris.data
    df_X = pd.DataFrame(
        X, columns=["sepal length", "sepal width", "petal length", "petal width"]
    )
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = EFS(
        knn,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        print_progress=False,
        feature_groups=[
            ["sepal length"],
            ["sepal width", "petal length"],
            ["petal width"],
        ],
    )
    efs1 = efs1.fit(df_X, y)
    expect = {
        0: {
            "feature_idx": (0, 1, 2),
            "feature_names": ("sepal length", "sepal width", "petal length"),
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.96]),
        },
        1: {
            "feature_idx": (0, 3),
            "feature_names": ("sepal length", "petal width"),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
        2: {
            "feature_idx": (1, 2, 3),
            "feature_names": ("sepal width", "petal length", "petal width"),
            "avg_score": 0.97333333333333338,
            "cv_scores": np.array([0.97333333]),
        },
    }
    dict_compare_utility(d1=expect, d2=efs1.subsets_)


def test_knn_wo_cv_with_fixed_features_and_feature_groups_case1():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = EFS(
        knn,
        min_features=1,
        max_features=2,
        scoring="accuracy",
        cv=0,
        print_progress=False,
        fixed_features=[0, 1],
        feature_groups=[[0, 1], [2], [3]],
    )
    efs1 = efs1.fit(X, y)
    # expect is based on what provided in `test_knn_wo_cv`
    expect = {
        0: {
            "feature_idx": (0, 1),
            "feature_names": ("0", "1"),
            "avg_score": 0.8333333333333334,
            "cv_scores": np.array([0.8333333333333334]),
        },
        1: {
            "feature_idx": (0, 1, 2),
            "feature_names": ("0", "1", "2"),
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.96]),
        },
        2: {
            "feature_idx": (0, 1, 3),
            "feature_names": ("0", "1", "3"),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
    }
    dict_compare_utility(d1=expect, d2=efs1.subsets_)


def test_knn_wo_cv_with_fixed_features_and_feature_groups_case2():
    # similar to case1, but `fixed_features` is now consisting of two groups
    # [0,1] and [3]
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = EFS(
        knn,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        print_progress=False,
        fixed_features=[0, 1, 3],
        feature_groups=[[0, 1], [2], [3]],
    )
    efs1 = efs1.fit(X, y)
    # expect is based on what provided in `test_knn_wo_cv`
    expect = {
        0: {
            "feature_idx": (0, 1, 3),
            "feature_names": ("0", "1", "3"),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
    }
    dict_compare_utility(d1=expect, d2=efs1.subsets_)


def test_check_support_string_in_feature_groups():
    iris = load_iris()
    X = iris.data
    y = iris.target

    features_names = ["sepal length", "sepal width", "petal length", "petal width"]
    df = pd.DataFrame(X, columns=features_names)

    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = EFS(
        knn,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        print_progress=False,
        feature_groups=[
            [features_names[0]],
            [features_names[1], features_names[2]],
            [features_names[3]],
        ],
    )
    efs1 = efs1.fit(df, y)
    # expect is based on what provided in `test_knn_wo_cv`
    expect = {
        0: {
            "feature_idx": (0, 1, 2),
            "feature_names": (features_names[0], features_names[1], features_names[2]),
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.96]),
        },
        1: {
            "feature_idx": (0, 3),
            "feature_names": (features_names[0], features_names[3]),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
        2: {
            "feature_idx": (1, 2, 3),
            "feature_names": (features_names[1], features_names[2], features_names[3]),
            "avg_score": 0.97333333333333338,
            "cv_scores": np.array([0.97333333]),
        },
    }
    dict_compare_utility(d1=expect, d2=efs1.subsets_)


def test_check_support_string_in_fixed_feature():
    iris = load_iris()
    X = iris.data
    y = iris.target

    features_names = ["sepal length", "sepal width", "petal length", "petal width"]
    df = pd.DataFrame(X, columns=features_names)

    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = EFS(
        knn,
        min_features=2,
        max_features=3,
        scoring="accuracy",
        cv=0,
        print_progress=False,
        fixed_features=[features_names[0], features_names[1]],
    )

    efs1 = efs1.fit(df, y)
    # expect is based on what provided in `test_knn_wo_cv`
    expect = {
        0: {
            "feature_idx": (0, 1),
            "feature_names": (features_names[0], features_names[1]),
            "avg_score": 0.8333333333333334,
            "cv_scores": np.array([0.8333333333333334]),
        },
        1: {
            "feature_idx": (0, 1, 2),
            "feature_names": (features_names[0], features_names[1], features_names[2]),
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.96]),
        },
        2: {
            "feature_idx": (0, 1, 3),
            "feature_names": (features_names[0], features_names[1], features_names[3]),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
    }
    dict_compare_utility(d1=expect, d2=efs1.subsets_)


def test_fixed_features_and_feature_groups_pandas_and_strings():
    iris = load_iris()
    X = iris.data
    feature_names = ["sepal length", "sepal width", "petal length", "petal width"]
    df_X = pd.DataFrame(X, columns=feature_names)
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = EFS(
        knn,
        min_features=2,
        max_features=2,
        scoring="accuracy",
        cv=0,
        print_progress=False,
        fixed_features=[feature_names[0], feature_names[1], feature_names[3]],
        feature_groups=[
            [feature_names[0], feature_names[1]],
            [feature_names[2]],
            [feature_names[3]],
        ],
    )
    efs1 = efs1.fit(df_X, y)
    # expect is based on what provided in `test_knn_wo_cv`
    expect = {
        0: {
            "feature_idx": (0, 1, 3),
            "feature_names": (feature_names[0], feature_names[1], feature_names[3]),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
    }
    dict_compare_utility(d1=expect, d2=efs1.subsets_)



# Out-of-fold stacking regressor tests
#
# Sebastian Raschka 2014-2024
#
# mlxtend Machine Learning Library Extensions
# Author: Eike Dehling <e.e.dehling@gmail.com>
#
# License: BSD 3 clause

import random

import numpy as np
import pytest
from packaging.version import Version
from scipy import sparse
from sklearn import __version__ as sklearn_version
from sklearn.base import clone
from sklearn.linear_model import Lasso, LinearRegression, Ridge
from sklearn.model_selection import GridSearchCV, KFold, train_test_split
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR

from mlxtend.externals.estimator_checks import NotFittedError
from mlxtend.regressor import StackingCVRegressor
from mlxtend.utils import assert_raises

# Some test data
np.random.seed(1)
X1 = np.sort(5 * np.random.rand(40, 1), axis=0)
X2 = np.sort(5 * np.random.rand(40, 2), axis=0)
X3 = np.zeros((40, 3))
y = np.sin(X1).ravel()
y[::5] += 3 * (0.5 - np.random.rand(8))
y2 = np.sin(X2)
y3 = np.zeros((40,))


def test_different_models():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingCVRegressor(
        regressors=[svr_lin, lr, ridge], meta_regressor=svr_rbf, random_state=0
    )
    stack.fit(X1, y).predict(X1)
    mse = 0.20
    got = np.mean((stack.predict(X1) - y) ** 2)
    assert round(got, 2) == mse, got


def test_use_features_in_secondary():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingCVRegressor(
        regressors=[svr_lin, lr, ridge],
        meta_regressor=svr_rbf,
        cv=3,
        random_state=0,
        use_features_in_secondary=True,
    )
    stack.fit(X1, y).predict(X1)
    mse = 0.2
    got = np.mean((stack.predict(X1) - y) ** 2)
    assert round(got, 2) == mse, "%f != %f" % (round(got, 2), mse)


def test_multivariate():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingCVRegressor(
        regressors=[svr_lin, lr, ridge], meta_regressor=svr_rbf, random_state=0
    )
    stack.fit(X2, y).predict(X2)
    mse = 0.20
    got = np.mean((stack.predict(X2) - y) ** 2)
    assert round(got, 2) == mse, "%f != %f" % (round(got, 2), mse)


def test_multivariate_class():
    lr = LinearRegression()
    ridge = Ridge(random_state=1)
    meta = LinearRegression()
    stregr = StackingCVRegressor(
        regressors=[lr, ridge], meta_regressor=meta, multi_output=True, random_state=0
    )
    stregr.fit(X2, y2).predict(X2)
    mse = 0.13
    got = np.mean((stregr.predict(X2) - y2) ** 2.0)
    assert round(got, 2) == mse, got


def test_internals():
    lr = LinearRegression()
    regressors = [lr, lr, lr, lr, lr]
    cv = 10
    stack = StackingCVRegressor(
        regressors=[lr, lr, lr, lr, lr], meta_regressor=lr, cv=cv, random_state=0
    )
    stack.fit(X3, y3)
    assert stack.predict(X3).mean() == y3.mean()
    assert stack.meta_regr_.intercept_ == 0.0
    assert stack.meta_regr_.coef_[0] == 0.0
    assert stack.meta_regr_.coef_[1] == 0.0
    assert stack.meta_regr_.coef_[2] == 0.0
    assert len(stack.regr_) == len(regressors)


def test_gridsearch_numerate_regr():
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingCVRegressor(
        regressors=[svr_lin, ridge, ridge], meta_regressor=svr_rbf, random_state=42
    )

    params = {
        "ridge-1__alpha": [0.01, 1.0],
        "ridge-2__alpha": [0.01, 1.0],
        "svr__C": [0.01, 1.0],
        "meta_regressor__C": [0.01, 1.0],
        "use_features_in_secondary": [True, False],
    }

    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(
            estimator=stack, param_grid=params, cv=5, iid=False, refit=True, verbose=0
        )
    else:
        grid = GridSearchCV(
            estimator=stack, param_grid=params, cv=5, refit=True, verbose=0
        )
    grid = grid.fit(X1, y)
    got = round(grid.best_score_, 1)
    assert got >= 0.1 and got <= 0.2, "%f is wrong" % got


def test_get_params():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf")
    ridge = Ridge(random_state=1)
    stregr = StackingCVRegressor(
        regressors=[ridge, lr], meta_regressor=svr_rbf, random_state=42
    )

    got = sorted(list({s.split("__")[0] for s in stregr.get_params().keys()}))
    expect = [
        "cv",
        "linearregression",
        "meta_regressor",
        "multi_output",
        "n_jobs",
        "pre_dispatch",
        "random_state",
        "refit",
        "regressors",
        "ridge",
        "shuffle",
        "store_train_meta_features",
        "use_features_in_secondary",
        "verbose",
    ]
    assert got == expect, got


def test_regressor_gridsearch():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingCVRegressor(
        regressors=[lr], meta_regressor=svr_rbf, random_state=1
    )

    params = {"regressors": [[ridge, lr], [lr, ridge, lr]]}

    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(
            estimator=stregr, param_grid=params, iid=False, cv=5, refit=True
        )
    else:
        grid = GridSearchCV(estimator=stregr, param_grid=params, cv=5, refit=True)
    grid.fit(X1, y)

    assert len(grid.best_params_["regressors"]) == 3


def test_predict_meta_features():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingCVRegressor(regressors=[lr, ridge], meta_regressor=svr_rbf)
    X_train, X_test, y_train, y_test = train_test_split(X2, y, test_size=0.3)
    stregr.fit(X_train, y_train)
    test_meta_features = stregr.predict(X_test)
    assert test_meta_features.shape[0] == X_test.shape[0]


def test_train_meta_features_():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingCVRegressor(
        regressors=[lr, ridge], meta_regressor=svr_rbf, store_train_meta_features=True
    )
    X_train, X_test, y_train, y_test = train_test_split(X2, y, test_size=0.3)
    stregr.fit(X_train, y_train)
    train_meta_features = stregr.train_meta_features_
    assert train_meta_features.shape[0] == X_train.shape[0]


def test_not_fitted_predict():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingCVRegressor(
        regressors=[lr, ridge], meta_regressor=svr_rbf, store_train_meta_features=True
    )
    X_train, X_test, y_train, y_test = train_test_split(X2, y, test_size=0.3)

    expect = (
        "This StackingCVRegressor instance is not fitted yet. Call "
        "'fit' with appropriate arguments before using this method."
    )

    assert_raises(NotFittedError, expect, stregr.predict, X_train)

    assert_raises(NotFittedError, expect, stregr.predict_meta_features, X_train)


def test_clone():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingCVRegressor(
        regressors=[lr, ridge], meta_regressor=svr_rbf, store_train_meta_features=True
    )
    clone(stregr)


def test_sparse_matrix_inputs():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingCVRegressor(
        regressors=[svr_lin, lr, ridge], meta_regressor=svr_rbf, random_state=42
    )

    # dense
    stack.fit(X1, y).predict(X1)
    mse = 0.21
    got = np.mean((stack.predict(X1) - y) ** 2)
    assert round(got, 2) == mse, got

    # sparse
    stack.fit(sparse.csr_matrix(X1), y)

    if Version(sklearn_version) < Version("0.21"):
        expected_value = 0.20
    elif Version(sklearn_version) < Version("0.22"):
        expected_value = 0.20
    else:
        expected_value = 0.21

    got = np.mean((stack.predict(sparse.csr_matrix(X1)) - y) ** 2)
    assert round(got, 2) == expected_value, got


def test_sparse_matrix_inputs_with_features_in_secondary():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingCVRegressor(
        regressors=[svr_lin, lr, ridge],
        meta_regressor=svr_rbf,
        random_state=42,
        use_features_in_secondary=True,
    )

    # dense
    stack.fit(X1, y).predict(X1)
    mse = 0.20
    got = np.mean((stack.predict(X1) - y) ** 2)
    assert round(got, 2) == mse, got

    # sparse
    stack.fit(sparse.csr_matrix(X1), y)
    mse = 0.20
    got = np.mean((stack.predict(sparse.csr_matrix(X1)) - y) ** 2)
    assert round(got, 2) == mse, got


# Calling for np.random will break the existing tests by changing the
# seed for CV.
# As a temporary workaround, we use random package to generate random w.
random.seed(8)
w = np.array([random.random() for _ in range(40)])
# w  = np.random.random(40)


def test_sample_weight():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingCVRegressor(
        regressors=[svr_lin, lr, ridge],
        meta_regressor=svr_rbf,
        cv=KFold(4, shuffle=True, random_state=7),
    )
    pred1 = stack.fit(X1, y, sample_weight=w).predict(X1)
    mse = 0.21  # 0.20770
    got = np.mean((stack.predict(X1) - y) ** 2)
    assert round(got, 2) == mse, "Expected %.2f, but got %.5f" % (mse, got)
    pred2 = stack.fit(X1, y).predict(X1)
    maxdiff = np.max(np.abs(pred1 - pred2))
    assert maxdiff > 1e-3, "max diff is %.4f" % maxdiff


def test_weight_ones():
    # sample_weight = None and sample_weight = ones
    # should give the same result, provided that the
    # randomness of the models is controled
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingCVRegressor(
        regressors=[svr_lin, lr, ridge],
        meta_regressor=svr_rbf,
        cv=KFold(5, shuffle=True, random_state=5),
    )
    pred1 = stack.fit(X1, y).predict(X1)
    pred2 = stack.fit(X1, y, sample_weight=np.ones(40)).predict(X1)
    assert np.max(np.abs(pred1 - pred2)) < 1e-3


def test_unsupported_regressor():
    # including regressor that does not support
    # sample_weight should raise error
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    knn = KNeighborsRegressor()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingCVRegressor(
        regressors=[svr_lin, lr, ridge, knn], meta_regressor=svr_rbf
    )
    with pytest.raises(TypeError):
        stack.fit(X1, y, sample_weight=w).predict(X1)


def test_unsupported_meta_regressor():
    # meta regressor with no support for
    # sample_weight should raise error
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    knn = KNeighborsRegressor()
    stack = StackingCVRegressor(regressors=[svr_lin, lr, ridge], meta_regressor=knn)

    with pytest.raises(TypeError):
        stack.fit(X1, y, sample_weight=w).predict(X1)


def test_weight_unsupported_with_no_weight():
    # pass no weight to regressors with no weight support
    # should not be a problem
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    knn = KNeighborsRegressor()
    stack = StackingCVRegressor(regressors=[svr_lin, lr, knn], meta_regressor=ridge)
    stack.fit(X1, y).predict(X1)

    stack = StackingCVRegressor(regressors=[svr_lin, lr, ridge], meta_regressor=knn)
    stack.fit(X1, y).predict(X1)


def test_gridsearch_replace_mix():
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    lr = LinearRegression()
    lasso = Lasso(random_state=1)
    stack = StackingCVRegressor(
        regressors=[svr_lin, lasso, ridge], meta_regressor=svr_rbf, shuffle=False
    )

    params = {
        "regressors": [[svr_lin, lr]],
        "linearregression": [None, lasso, ridge],
        "svr__kernel": ["poly"],
    }

    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(
            estimator=stack,
            param_grid=params,
            cv=KFold(5, shuffle=True, random_state=42),
            iid=False,
            refit=True,
            verbose=0,
        )
    else:
        grid = GridSearchCV(
            estimator=stack,
            param_grid=params,
            cv=KFold(5, shuffle=True, random_state=42),
            refit=True,
            verbose=0,
        )
    grid = grid.fit(X1, y)

    got1 = round(grid.best_score_, 2)
    got2 = len(grid.best_params_["regressors"])
    got3 = grid.best_params_["regressors"][0].kernel

    assert got1 == 0.73, got1
    assert got2 == 2, got2
    assert got3 == "poly", got3



import sys

if sys.version_info < (3, 0):
    from nose.plugins.skip import SkipTest

    raise SkipTest

from mlxtend.text import generalize_names


def test_generalize_names():
    assert generalize_names("Samuel Eto'o") == "etoo s"
    assert generalize_names("Eto'o, Samuel") == "etoo s"
    assert generalize_names("Eto'o, Samuel") == "etoo s"
    assert generalize_names("Xavi") == "xavi"
    assert generalize_names("Yaya Toure") == "toure y"
    assert generalize_names("Pozo, Jose Angel") == "pozo j"
    assert generalize_names("Pozo, Jose Angel") == "pozo j"
    assert generalize_names("Jose Angel Pozo") == "pozo j"
    assert generalize_names("Jose Pozo") == "pozo j"
    assert generalize_names("Pozo, Jose Angel", firstname_output_letters=2) == "pozo jo"
    assert generalize_names("Eto'o, Samuel", firstname_output_letters=2) == "etoo sa"
    assert generalize_names("Eto'o, Samuel", firstname_output_letters=0) == "etoo"
    assert generalize_names("Eto'o, Samuel", output_sep=", ") == "etoo, s"
    assert generalize_names("Eto'o, Samuel", output_sep=", ") == "etoo, s"

    assert generalize_names("van Persie, Robin", output_sep=", ") == "vanpersie, r"
    assert generalize_names("Robin van Persie", output_sep=", ") == "vanpersie, r"
    assert generalize_names("Rafael van der Vaart", output_sep=", ") == "vandervaart, r"
    assert (
        generalize_names("van der Vaart, Rafael", output_sep=", ") == "vandervaart, r"
    )
    assert generalize_names("Ben Hamer") == "hamer b"



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from mlxtend._base import _BaseModel
from mlxtend.utils import assert_raises


class BlankModel(_BaseModel):
    def __init__(self, print_progress=0, random_seed=1):
        self.print_progress = print_progress
        self.random_seed = random_seed


def test_init():
    est = BlankModel(print_progress=0, random_seed=1)
    assert hasattr(est, "print_progress")
    assert hasattr(est, "random_seed")


def test_check_array_1():
    X = np.array([1, 2, 3])
    est = BlankModel()
    assert_raises(
        ValueError,
        "X must be a 2D array. Try X[:, numpy.newaxis]",
        est._check_arrays,
        X,
    )


def test_check_array_2():
    X = list([[1], [2], [3]])
    est = BlankModel(print_progress=0, random_seed=1)

    assert_raises(ValueError, "X must be a numpy array", est._check_arrays, X)


def test_check_array_3():
    X = np.array([[1], [2], [3]])
    est = BlankModel(print_progress=0, random_seed=1)
    est._check_arrays(X)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# A function for removing chart junk from matplotlib plots
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


def remove_borders(axes, left=False, bottom=False, right=True, top=True):
    """Remove chart junk from matplotlib plots.

    Parameters
    ----------
    axes : iterable
        An iterable containing plt.gca()
        or plt.subplot() objects, e.g. [plt.gca()].
    left : bool (default: `False`)
        Hide left axis spine if True.
    bottom : bool (default: `False`)
        Hide bottom axis spine if True.
    right : bool (default: `True`)
        Hide right axis spine if True.
    top : bool (default: `True`)
        Hide top axis spine if True.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/plotting/remove_chartjunk/

    """
    for ax in axes:
        ax.spines["top"].set_visible(not top)
        ax.spines["right"].set_visible(not right)
        ax.spines["bottom"].set_visible(not bottom)
        ax.spines["left"].set_visible(not left)
        if bottom:
            ax.tick_params(bottom="off", labelbottom="off")
        if top:
            ax.tick_params(top="off")
        if left:
            ax.tick_params(left="off", labelleft="off")
        if right:
            ax.tick_params(right="off")



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from mlxtend._base import _MultiClass


class BlankClassifier(_MultiClass):
    def __init__(self, print_progress=0, random_seed=1):
        self.print_progress = print_progress
        self.random_seed = random_seed


def test_default():
    y = np.array([0, 1, 2, 3, 4, 2])
    mc = BlankClassifier()
    expect = np.array(
        [
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0, 0.0, 0.0],
        ],
        dtype="float",
    )
    out = mc._one_hot(y=y, n_labels=5, dtype="float")
    np.testing.assert_array_equal(expect, out)


def test_oneclass():
    y = np.array([0, 0, 0])
    mc = BlankClassifier()
    out = mc._one_hot(y=y, n_labels=1, dtype="float")
    expect = np.array([[1.0], [1.0], [1.0]])
    np.testing.assert_array_equal(expect, out)


def test_morelabels():
    y = np.array([0, 0, 1])
    mc = BlankClassifier()
    out = mc._one_hot(y=y, n_labels=4, dtype="float")
    expect = np.array(
        [[1.0, 0.0, 0.0, 0.0], [1.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0]]
    )
    np.testing.assert_array_equal(expect, out)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from numpy.testing import assert_array_equal

from mlxtend.evaluate import confusion_matrix


def test_multiclass():
    y_targ = [1, 1, 1, 0, 0, 2, 0, 3]
    y_pred = [1, 0, 1, 0, 0, 2, 1, 3]
    x = np.array([[2, 1, 0, 0], [1, 2, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
    y = confusion_matrix(y_targ, y_pred, binary=False, positive_label=1)
    assert_array_equal(x, y)


def test_binary():
    y_targ = [1, 1, 1, 0, 0, 2, 0, 3]
    y_pred = [1, 0, 1, 0, 0, 2, 1, 3]
    x = np.array([[4, 1], [1, 2]])
    y = confusion_matrix(y_targ, y_pred, binary=True, positive_label=1)
    assert_array_equal(x, y)



import sys

if sys.version_info < (3, 0):
    from nose.plugins.skip import SkipTest

    raise SkipTest

from io import StringIO

import pandas as pd
from data_names import csv

from mlxtend.text import generalize_names, generalize_names_duplcheck


def test_generalize_names_duplcheck():
    df = pd.read_csv(StringIO(csv))

    # duplicates before
    dupl = any(df["Name"].apply(generalize_names).duplicated())
    assert dupl is True

    # no duplicates
    df_new = generalize_names_duplcheck(df=df, col_name="Name")
    no_dupl = any(df_new["Name"].duplicated())
    assert no_dupl is False



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import sys
from contextlib import contextmanager
from io import StringIO

import numpy as np
import pandas as pd
from numpy.testing import assert_array_equal
from packaging.version import Version
from pandas import __version__ as pandas_version
from scipy.sparse import csr_matrix

from mlxtend.preprocessing import TransactionEncoder
from mlxtend.utils import assert_raises


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class FPTestEdgeCases(object):
    """
    Base class for testing edge cases for pattern mining.
    """

    def setUp(self, fpalgo):
        self.fpalgo = fpalgo

    def test_empty(self):
        df = pd.DataFrame([[]])
        res_df = self.fpalgo(df)
        expect = pd.DataFrame([], columns=["support", "itemsets"])
        compare_dataframes(res_df, expect)


class FPTestErrors(object):
    """
    Base class for testing expected errors for pattern mining.
    """

    def setUp(self, fpalgo):
        self.one_ary = np.array(
            [
                [0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 1],
                [0, 0, 1, 1, 0, 1, 0, 1, 1, 0, 1],
                [1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1],
                [0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0],
            ]
        )
        self.cols = [
            "Apple",
            "Corn",
            "Dill",
            "Eggs",
            "Ice cream",
            "Kidney Beans",
            "Milk",
            "Nutmeg",
            "Onion",
            "Unicorn",
            "Yogurt",
        ]
        self.df = pd.DataFrame(self.one_ary, columns=self.cols)
        self.fpalgo = fpalgo

    def test_itemsets_type(self):
        # This is default behavior
        res_colindice = self.fpalgo(self.df, use_colnames=False)
        for i in res_colindice["itemsets"]:
            assert isinstance(i, frozenset) is True

        res_colnames = self.fpalgo(self.df, use_colnames=True)
        for i in res_colnames["itemsets"]:
            assert isinstance(i, frozenset) is True

    def test_raise_error_if_input_is_not_binary(self):
        def test_with_dataframe(df):
            assert_raises(
                ValueError,
                "The allowed values for a DataFrame are True, "
                "False, 0, 1. Found value 2",
                self.fpalgo,
                df,
            )

        df2 = pd.DataFrame(self.one_ary, columns=self.cols).copy()
        df2.iloc[3, 3] = 2
        test_with_dataframe(df2)

        sdf = df2.astype(pd.SparseDtype("int", fill_value=0))
        test_with_dataframe(sdf)

    def test_sparsedataframe_notzero_column(self):
        dfs = self.df.astype(pd.SparseDtype("int", 0))

        dfs.columns = [i for i in range(len(dfs.columns))]
        self.fpalgo(dfs)

        dfs = self.df.astype(pd.SparseDtype("int", 0))

        dfs.columns = [i + 1 for i in range(len(dfs.columns))]
        assert_raises(
            ValueError,
            "Due to current limitations in Pandas, "
            "if the sparse format has integer column names,"
            "names, please make sure they either start "
            "with `0` or cast them as string column names: "
            "`df.columns = [str(i) for i in df.columns`].",
            self.fpalgo,
            dfs,
        )


class FPTestEx1(object):
    """
    Base class for testing frequent pattern mining on a small example.
    """

    def setUp(self, fpalgo, one_ary=None):
        if one_ary is None:
            self.one_ary = np.array(
                [
                    [0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 1],
                    [0, 0, 1, 1, 0, 1, 0, 1, 1, 0, 1],
                    [1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0],
                    [0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1],
                    [0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0],
                ]
            )
        else:
            self.one_ary = one_ary

        self.cols = [
            "Apple",
            "Corn",
            "Dill",
            "Eggs",
            "Ice cream",
            "Kidney Beans",
            "Milk",
            "Nutmeg",
            "Onion",
            "Unicorn",
            "Yogurt",
        ]

        self.df = pd.DataFrame(self.one_ary, columns=self.cols)

        self.fpalgo = fpalgo

    def test_frozenset_selection(self):
        res_df = self.fpalgo(self.df, use_colnames=True)
        assert res_df.values.shape == self.fpalgo(self.df).values.shape
        assert res_df[res_df["itemsets"] == "nothing"].values.shape == (0, 2)
        assert res_df[res_df["itemsets"] == {"Milk", "Kidney Beans"}].values.shape == (
            1,
            2,
        )
        assert res_df[
            res_df["itemsets"] == frozenset(("Milk", "Kidney Beans"))
        ].values.shape == (1, 2)
        assert res_df[
            res_df["itemsets"] == frozenset(("Kidney Beans", "Milk"))
        ].values.shape == (1, 2)

    def test_sparse(self):
        def test_with_fill_values(fill_value):
            sdt = pd.SparseDtype(type(fill_value), fill_value=fill_value)
            sdf = self.df.astype(sdt)
            res_df = self.fpalgo(sdf, use_colnames=True)
            assert res_df.values.shape == self.fpalgo(self.df).values.shape
            assert res_df[res_df["itemsets"] == "nothing"].values.shape == (0, 2)
            assert res_df[
                res_df["itemsets"] == {"Milk", "Kidney Beans"}
            ].values.shape == (1, 2)
            assert res_df[
                res_df["itemsets"] == frozenset(("Milk", "Kidney Beans"))
            ].values.shape == (1, 2)
            assert res_df[
                res_df["itemsets"] == frozenset(("Kidney Beans", "Milk"))
            ].values.shape == (1, 2)

        test_with_fill_values(0)
        test_with_fill_values(False)

    def test_sparse_with_zero(self):
        if Version(pandas_version) < Version("1.2"):
            # needs to be revisited in future when pandas bug
            # in https://github.com/pandas-dev/pandas/issues/29814
            # is fixed
            return
        res_df = self.fpalgo(self.df)
        ary2 = self.one_ary.copy()
        ary2[3, :] = 1
        sparse_ary = csr_matrix(ary2)
        sparse_ary[3, :] = self.one_ary[3, :]

        sdf = pd.DataFrame.sparse.from_spmatrix(sparse_ary, columns=self.df.columns)
        res_df2 = self.fpalgo(sdf)
        compare_dataframes(res_df2, res_df)


class FPTestEx1All(FPTestEx1):
    def setUp(self, fpalgo, one_ary=None):
        FPTestEx1.setUp(self, fpalgo, one_ary=one_ary)

    def test_default(self):
        res_df = self.fpalgo(self.df)
        expect = pd.DataFrame(
            [
                [0.8, np.array([3])],
                [1.0, np.array([5])],
                [0.6, np.array([6])],
                [0.6, np.array([8])],
                [0.6, np.array([10])],
                [0.8, np.array([3, 5])],
                [0.6, np.array([3, 8])],
                [0.6, np.array([5, 6])],
                [0.6, np.array([5, 8])],
                [0.6, np.array([5, 10])],
                [0.6, np.array([3, 5, 8])],
            ],
            columns=["support", "itemsets"],
        )

        compare_dataframes(res_df, expect)

    def test_max_len(self):
        res_df1 = self.fpalgo(self.df)
        max_len = np.max(res_df1["itemsets"].apply(len))
        assert max_len == 3

        res_df2 = self.fpalgo(self.df, max_len=2)
        max_len = np.max(res_df2["itemsets"].apply(len))
        assert max_len == 2

    def test_low_memory_flag(self):
        import inspect

        if "low_memory" in inspect.signature(self.fpalgo).parameters:
            with captured_output() as (out, err):
                _ = self.fpalgo(self.df, low_memory=True, verbose=1)

            # Only get the last value of the stream to reduce test noise
            expect = "Processing 4 combinations | Sampling itemset size 3\n"
            out = out.getvalue().split("\r")[-1]
            assert out == expect
        else:
            # If there is no low_memory argument, don't run the test.
            assert True


class FPTestEx2(object):
    """
    Base class for testing frequent pattern mining on a small example.
    """

    def setUp(self):
        database = [["a"], ["b"], ["c", "d"], ["e"]]
        te = TransactionEncoder()
        te_ary = te.fit(database).transform(database)

        self.df = pd.DataFrame(te_ary, columns=te.columns_)


class FPTestEx2All(FPTestEx2):
    def setUp(self, fpalgo):
        self.fpalgo = fpalgo
        FPTestEx2.setUp(self)

    def test_output(self):
        res_df = self.fpalgo(self.df, min_support=0.001, use_colnames=True)
        expect = pd.DataFrame(
            [
                [0.25, frozenset(["a"])],
                [0.25, frozenset(["b"])],
                [0.25, frozenset(["c"])],
                [0.25, frozenset(["d"])],
                [0.25, frozenset(["e"])],
                [0.25, frozenset(["c", "d"])],
            ],
            columns=["support", "itemsets"],
        )

        compare_dataframes(res_df, expect)


class FPTestEx3(object):
    """
    Base class for testing frequent pattern mining on a small example.
    """

    def setUp(self):
        database = [["a"], ["b"], ["c", "d"], ["e"]]
        te = TransactionEncoder()
        te_ary = te.fit(database).transform(database)

        self.df = pd.DataFrame(te_ary, columns=te.columns_)


class FPTestEx3All(FPTestEx3):
    def setUp(self, fpalgo):
        self.fpalgo = fpalgo
        FPTestEx3.setUp(self)

    def test_output3(self):
        assert_raises(
            ValueError,
            "`min_support` must be a positive "
            "number within the interval `(0, 1]`. Got 0.0.",
            self.fpalgo,
            self.df,
            min_support=0.0,
        )


def compare_dataframes(df1, df2):
    itemsets1 = [sorted(list(i)) for i in df1["itemsets"]]
    itemsets2 = [sorted(list(i)) for i in df2["itemsets"]]

    rows1 = sorted(zip(itemsets1, df1["support"]))
    rows2 = sorted(zip(itemsets2, df2["support"]))

    for row1, row2 in zip(rows1, rows2):
        if row1[0] != row2[0]:
            msg = f"Expected different frequent itemsets\nx:{row1[0]}\ny:{row2[0]}"
            raise AssertionError(msg)
        elif row1[1] != row2[1]:
            msg = f"Expected different support\nx:{row1[1]}\ny:{row2[1]}"
            raise AssertionError(msg)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import sys

import numpy as np
from sklearn.base import clone

from mlxtend.classifier import Perceptron
from mlxtend.data import iris_data
from mlxtend.utils import assert_raises

# Iris Data
X, y = iris_data()
X = X[:, [0, 3]]  # sepal length and petal width
X = X[0:100]  # class 0 and class 1
y0 = y[0:100]  # class 0 and class 1

# standardize
X_std = np.copy(X)
X_std[:, 0] = (X[:, 0] - X[:, 0].mean()) / X[:, 0].std()
X_std[:, 1] = (X[:, 1] - X[:, 1].mean()) / X[:, 1].std()


def test_invalid_labels_1():
    y1 = np.where(y0 == 0, 2, 1)
    ppn = Perceptron(epochs=15, eta=0.01, random_seed=1)

    if sys.version_info >= (3, 0):
        objtype = "{(0, 1)}"
    else:
        objtype = "set([(0, 1)])"

    expect = "Labels not in %s.\nFound (1, 2)" % objtype

    assert_raises(AttributeError, expect, ppn.fit, X, y1, {(0, 1)})


def test_invalid_labels_2():
    y1 = np.where(y0 == 0, -1, 1)
    ppn = Perceptron(epochs=15, eta=0.01, random_seed=1)

    assert_raises(
        AttributeError,
        "y array must not contain negative labels.\nFound [-1  1]",
        ppn.fit,
        X,
        y1,
        {(-1, 1)},
    )


def test_standardized_iris_data():
    ppn = Perceptron(epochs=15, eta=0.01, random_seed=1)
    ppn = ppn.fit(X_std, y0)
    assert (y0 == ppn.predict(X_std)).all(), ppn.predict(X_std)


def test_progress_1():
    ppn = Perceptron(epochs=15, eta=0.01, random_seed=1, print_progress=1)
    ppn = ppn.fit(X_std, y0)


def test_progress_2():
    ppn = Perceptron(epochs=15, eta=0.01, random_seed=1, print_progress=2)
    ppn = ppn.fit(X_std, y0)


def test_progress_3():
    ppn = Perceptron(epochs=15, eta=0.01, random_seed=1, print_progress=3)
    ppn = ppn.fit(X_std, y0)


def test_score_function():
    ppn = Perceptron(epochs=15, eta=0.01, random_seed=1)
    ppn = ppn.fit(X_std, y0)
    acc = ppn.score(X_std, y0)
    assert acc == 1.0, acc


def test_nonstandardized_iris_data():
    ppn = Perceptron(epochs=100, eta=0.01, random_seed=1)
    ppn = ppn.fit(X, y0)
    assert (y0 == ppn.predict(X)).all()


def test_clone():
    ppn = Perceptron()
    clone(ppn)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from mlxtend.evaluate import mcnemar_table
from mlxtend.utils import assert_raises


def test_input_array_1d():
    t = np.array([[1, 2], [3, 4]])
    assert_raises(
        ValueError,
        "One or more input arrays are not 1-dimensional.",
        mcnemar_table,
        t,
        t,
        t,
    )


def test_input_array_lengths_1():
    t = np.array([1, 2])
    t2 = np.array([1, 2, 3])
    assert_raises(
        ValueError,
        ("y_target and y_model1 contain a different number" " of elements."),
        mcnemar_table,
        t,
        t2,
        t,
    )


def test_input_array_lengths_2():
    t = np.array([1, 2])
    t2 = np.array([1, 2, 3])
    assert_raises(
        ValueError,
        ("y_target and y_model2 contain a different number" " of elements."),
        mcnemar_table,
        t,
        t,
        t2,
    )


def test_input_binary_all_right():
    y_target = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    y_model1 = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    y_model2 = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    tb = mcnemar_table(y_target=y_target, y_model1=y_model1, y_model2=y_model2)
    expect = np.array([[8, 0], [0, 0]])
    np.testing.assert_array_equal(tb, expect)


def test_input_binary_all_wrong():
    y_target = np.array([1, 1, 1, 1, 0, 0, 0, 0])
    y_model1 = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    y_model2 = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    tb = mcnemar_table(y_target=y_target, y_model1=y_model1, y_model2=y_model2)
    expect = np.array([[0, 0], [0, 8]])

    np.testing.assert_array_equal(tb, expect)


def test_input_binary():
    y_target = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
    y_model1 = np.array([0, 1, 0, 0, 0, 1, 1, 0, 0, 0])
    y_model2 = np.array([0, 0, 1, 1, 0, 1, 1, 0, 0, 0])
    tb = mcnemar_table(y_target=y_target, y_model1=y_model1, y_model2=y_model2)
    expect = np.array([[4, 2], [1, 3]])

    np.testing.assert_array_equal(tb, expect)


def test_input_nonbinary():
    y_target = np.array([0, 0, 0, 0, 0, 2, 1, 1, 1, 1])
    y_model1 = np.array([0, 5, 0, 0, 0, 2, 1, 0, 0, 0])
    y_model2 = np.array([0, 0, 1, 3, 0, 2, 1, 0, 0, 0])

    tb = mcnemar_table(y_target=y_target, y_model1=y_model1, y_model2=y_model2)
    expect = np.array([[4, 2], [1, 3]])

    np.testing.assert_array_equal(tb, expect)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Principal Component Analysis for dimensionality reduction.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from scipy.spatial import distance

from .._base import _BaseModel


class RBFKernelPCA(_BaseModel):
    """
    RBF Kernel Principal Component Analysis for dimensionality reduction.

    Parameters
    ----------
    gamma : float (default: 15.0)
        Free parameter (coefficient) of the RBF kernel.
    n_components : int (default: None)
        The number of principal components for transformation.
        Keeps the original dimensions of the dataset if `None`.
    copy_X : bool (default: True)
        Copies training data, which is required to compute the projection
        of new data via the transform method. Uses a reference to X if False.

    Attributes
    ----------
    e_vals_ : array-like, shape=[n_features]
        Eigenvalues in sorted order.
    e_vecs_ : array-like, shape=[n_features]
       Eigenvectors in sorted order.
    X_projected_ : array-like, shape=[n_samples, n_components]
       Training samples projected along the component axes.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/feature_extraction/RBFKernelPCA/

    """

    def __init__(self, gamma=15.0, n_components=None, copy_X=True):
        if n_components is not None and n_components < 1:
            raise AttributeError("n_components must be > 1 or None")
        self.n_components = n_components
        self.gamma = gamma
        self.copy_X = copy_X
        self._is_fitted = False

    def fit(self, X):
        """Learn model from training data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        self : object

        """
        self._is_fitted = False
        self._check_arrays(X=X)
        self._fit(X=X)
        self._is_fitted = True
        return self

    def _fit(self, X):
        if self.n_components is None or self.n_components > X.shape[1]:
            n_components = X.shape[1]
        else:
            n_components = self.n_components
        kernel_mat = self._kernel_matrix(X=X, gamma=self.gamma)
        self.e_vals_, self.e_vecs_ = self._eigendecom(kernel_mat)
        self.X_projected_ = self._projection_matrix(
            eig_vecs=self.e_vecs_, n_components=n_components
        )
        if self.copy_X:
            self.X_ = X.copy()
        else:
            self.X_ = X
        return self

    def transform(self, X):
        """Apply the non-linear transformation on X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_projected : np.ndarray, shape = [n_samples, n_components]
            Projected training vectors.

        """
        if not hasattr(self, "X_"):
            raise AttributeError("Object as not been fitted, yet.")
        self._check_arrays(X=X)
        # pair_dist = np.array([np.sum((X - row)**2) for row in self.X_])
        pair_dist = np.ones((self.X_.shape[0], X.shape[0]))
        for idx in range(X.shape[0]):
            pair_dist[:, idx] = ((self.X_ - X[idx]) ** 2).sum(axis=1)

        K = np.exp((-1) * self.gamma * pair_dist)
        e_vecs = self._projection_matrix(
            eig_vecs=self.e_vecs_, n_components=self.n_components
        )
        return K.T.dot(e_vecs / self.e_vals_[: e_vecs.shape[1]])

    def _kernel_matrix(self, X, gamma):
        # Calculating the squared Euclidean distances for every pair of points
        # in the MxN dimensional dataset.
        sq_dists = distance.pdist(X, "sqeuclidean")

        # Converting the pairwise distances into a symmetric MxM matrix.
        mat_sq_dists = distance.squareform(sq_dists)

        # Computing the MxM kernel matrix.
        K = np.exp((-1) * gamma * mat_sq_dists)

        # Centering the symmetric NxN kernel matrix.
        N = K.shape[0]
        one_n = np.ones((N, N)) / N
        K = K - one_n.dot(K) - K.dot(one_n) + one_n.dot(K).dot(one_n)
        return K

    def _eigendecom(self, kernel_mat):
        e_vals, e_vecs = np.linalg.eigh(kernel_mat)
        sort_idx = np.argsort(e_vals)[::-1]
        e_vals, e_vecs = e_vals[sort_idx], e_vecs[:, sort_idx]
        return e_vals, e_vecs

    def _projection_matrix(self, eig_vecs, n_components):
        matrix_w = np.vstack([eig_vecs[:, i] for i in range(n_components)]).T
        return matrix_w



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import unittest

import numpy as np
from test_fpbase import (
    FPTestEdgeCases,
    FPTestErrors,
    FPTestEx1All,
    FPTestEx2All,
    FPTestEx3All,
)

from mlxtend.frequent_patterns import apriori


def apriori_wrapper_low_memory(*args, **kwargs):
    return apriori(*args, **kwargs, low_memory=True)


class TestEdgeCases(unittest.TestCase, FPTestEdgeCases):
    def setUp(self):
        FPTestEdgeCases.setUp(self, apriori)


class TestErrors(unittest.TestCase, FPTestErrors):
    def setUp(self):
        FPTestErrors.setUp(self, apriori)


class TestApriori(unittest.TestCase, FPTestEx1All):
    def setUp(self):
        FPTestEx1All.setUp(self, apriori)


class TestAprioriLowMemory(unittest.TestCase, FPTestEx1All):
    def setUp(self):
        FPTestEx1All.setUp(self, apriori_wrapper_low_memory)


class TestAprioriBoolInput(unittest.TestCase, FPTestEx1All):
    def setUp(self):
        one_ary = np.array(
            [
                [False, False, False, True, False, True, True, True, True, False, True],
                [False, False, True, True, False, True, False, True, True, False, True],
                [
                    True,
                    False,
                    False,
                    True,
                    False,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    False,
                    True,
                    False,
                    False,
                    False,
                    True,
                    True,
                    False,
                    False,
                    True,
                    True,
                ],
                [
                    False,
                    True,
                    False,
                    True,
                    True,
                    True,
                    False,
                    False,
                    True,
                    False,
                    False,
                ],
            ]
        )
        FPTestEx1All.setUp(self, apriori, one_ary=one_ary)


class TestEx2(unittest.TestCase, FPTestEx2All):
    def setUp(self):
        FPTestEx2All.setUp(self, apriori)


class TestEx3(unittest.TestCase, FPTestEx3All):
    def setUp(self):
        FPTestEx3All.setUp(self, apriori)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import sys

import numpy as np
from packaging.version import Version
from scipy.sparse import issparse
from sklearn import __version__ as sklearn_version
from sklearn.datasets import load_iris
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from mlxtend.preprocessing import CopyTransformer
from mlxtend.utils import assert_raises

iris = load_iris()
X, y = iris.data, iris.target


def test_copy():
    copy = CopyTransformer()
    np.testing.assert_array_equal(X, copy.transform(X))


def test_copy_failtype():
    copy = CopyTransformer()

    expect = (
        "X must be a list or NumPy array or SciPy sparse array." " Found <class 'int'>"
    )
    if sys.version_info < (3, 0):
        expect = expect.replace("class", "type")
    assert_raises(ValueError, expect, copy.transform, 1)


def test_pipeline():
    param_grid = [{"logisticregression__C": [1, 0.1, 10]}]
    pipe = make_pipeline(
        StandardScaler(),
        CopyTransformer(),
        LogisticRegression(solver="liblinear", multi_class="ovr"),
    )

    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(pipe, param_grid, cv=3, n_jobs=1, iid=False)
    else:
        grid = GridSearchCV(pipe, param_grid, cv=3, n_jobs=1)
    grid.fit(X, y)


def test_sparse():
    copy = CopyTransformer()
    tfidf = TfidfTransformer()
    X_t = tfidf.fit_transform([[1, 2, 3]])
    assert issparse(X_t)
    X_dense = copy.transform(X_t).toarray()
    expect = np.array([[0.26726124, 0.53452248, 0.80178373]])
    assert np.allclose(X_dense, expect)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from mlxtend.data import iris_data
from mlxtend.feature_extraction import LinearDiscriminantAnalysis as LDA
from mlxtend.preprocessing import standardize

X, y = iris_data()
X = standardize(X)


def test_default_components():
    lda = LDA()
    lda.fit(X, y)
    res = lda.fit(X, y).transform(X)
    assert res.shape[1] == 4


def test_default_2components():
    lda = LDA(n_discriminants=2)
    lda.fit(X, y)
    res = lda.fit(X, y).transform(X)
    assert res.shape[1] == 2


def test_default_0components():
    with pytest.raises(AttributeError):
        LDA(n_discriminants=0)


def test_evals():
    lda = LDA(n_discriminants=2)
    lda.fit(X, y).transform(X)
    np.set_printoptions(suppress=True)
    print("%s" % lda.e_vals_)
    assert_almost_equal(lda.e_vals_, [20.90, 0.14, 0.0, 0.0], decimal=2)


def test_fail_array_fit():
    lda = LDA()
    with pytest.raises(ValueError):
        lda.fit(X[1], y[1])


def test_fail_array_transform():
    lda = LDA()
    lda.fit(X, y)
    with pytest.raises(ValueError):
        lda.transform(X[1])



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause
import numpy as np
from numpy.testing import assert_almost_equal
from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LinearRegression
from sklearn.metrics import roc_auc_score
from sklearn.neighbors import KNeighborsClassifier

from mlxtend.data import boston_housing_data
from mlxtend.feature_selection import SequentialFeatureSelector as SFS
from mlxtend.utils import assert_raises


def dict_compare_utility(d_actual, d_desired, decimal=2):
    assert d_actual.keys() == d_desired.keys(), "%s != %s" % (d_actual, d_desired)
    for i in d_actual:
        err_msg = "d_actual[%s]['feature_idx']" " != d_desired[%s]['feature_idx']" % (
            i,
            i,
        )
        assert d_actual[i]["feature_idx"] == d_desired[i]["feature_idx"], err_msg
        assert_almost_equal(
            actual=d_actual[i]["avg_score"],
            desired=d_desired[i]["avg_score"],
            decimal=decimal,
            err_msg=(
                "d_actual[%s]['avg_score']" " != d_desired[%s]['avg_score']" % (i, i)
            ),
        )
        assert_almost_equal(
            actual=d_actual[i]["cv_scores"],
            desired=d_desired[i]["cv_scores"],
            decimal=decimal,
            err_msg=(
                "d_actual[%s]['cv_scores']" " != d_desired[%s]['cv_scores']" % (i, i)
            ),
        )


def test_run_default():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier()
    sfs = SFS(estimator=knn, verbose=0)
    sfs.fit(X, y)
    assert sfs.k_feature_idx_ == (3,)


def test_fit_params():
    iris = load_iris()
    X = iris.data
    y = iris.target
    sample_weight = np.ones(X.shape[0])
    forest = RandomForestClassifier(n_estimators=100, random_state=123)
    sfs = SFS(estimator=forest, verbose=0)
    sfs.fit(X, y, sample_weight=sample_weight)
    assert sfs.k_feature_idx_ == (3,)


def test_knn_wo_cv_feature_groups_default():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs1 = SFS(
        knn,
        k_features=3,
        forward=True,
        floating=False,
        cv=0,
        verbose=0,
        feature_groups=[[0], [1], [2], [3]],
    )
    sfs1 = sfs1.fit(X, y)
    expect = {
        1: {
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.96]),
            "feature_idx": (3,),
        },
        2: {
            "avg_score": 0.97333333333333338,
            "cv_scores": np.array([0.97333333]),
            "feature_idx": (2, 3),
        },
        3: {
            "avg_score": 0.97333333333333338,
            "cv_scores": np.array([0.97333333]),
            "feature_idx": (1, 2, 3),
        },
    }
    dict_compare_utility(d_actual=sfs1.subsets_, d_desired=expect, decimal=2)


def test_regression_sbfs():
    X, y = boston_housing_data()
    lr = LinearRegression()
    sfs_r = SFS(
        lr,
        k_features=1,  # this is number of groups if `feature_groups` is not None
        forward=False,
        floating=True,
        scoring="neg_mean_squared_error",
        cv=10,
        verbose=0,
        feature_groups=[[7, 10, 12], [0], [1], [2], [3], [4], [5], [6], [8], [9], [11]],
    )
    sfs_r = sfs_r.fit(X, y)
    assert sfs_r.k_feature_idx_ == (7, 10, 12), sfs_r.k_feature_idx_


def test_transform_not_fitted():
    iris = load_iris()
    X = iris.data
    knn = KNeighborsClassifier(n_neighbors=4)

    sfs1 = SFS(
        knn,
        k_features=2,
        forward=True,
        floating=False,
        cv=0,
        clone_estimator=False,
        verbose=0,
        n_jobs=1,
    )

    expect = "SequentialFeatureSelector has not been fitted, yet."

    assert_raises(AttributeError, expect, sfs1.transform, X)


def test_keyboard_interrupt():
    iris = load_iris()
    X = iris.data
    y = iris.target

    knn = KNeighborsClassifier(n_neighbors=4)
    sfs1 = SFS(
        knn,
        k_features=2,
        forward=True,
        floating=False,
        cv=3,
        clone_estimator=False,
        verbose=5,
        n_jobs=1,
        feature_groups=[[0, 1], [2], [3]],
    )

    sfs1._TESTING_INTERRUPT_MODE = True
    out = sfs1.fit(X, y)

    assert len(out.subsets_.keys()) > 0
    assert sfs1.interrupted_


def test_max_feature_subset_best():
    X, y = boston_housing_data()
    lr = LinearRegression()

    sfs = SFS(
        lr,
        k_features="best",
        forward=True,
        floating=False,
        cv=10,
        feature_groups=[
            [0],
            [2, 4],
            [1, 3, 5],
            [6],
            [7, 8, 9, 10],
            [11],
            [12],
        ],
    )

    sfs = sfs.fit(X, y)
    assert sfs.k_feature_idx_ == (1, 3, 5, 7, 8, 9, 10, 11, 12)


def test_max_feature_subset_parsimonious():
    X, y = boston_housing_data()
    lr = LinearRegression()

    sfs = SFS(
        lr,
        k_features="parsimonious",
        forward=True,
        floating=False,
        cv=10,
        feature_groups=[
            [0],
            [1, 3],
            [2, 4],
            [5, 10, 11, 12],
            [6],
            [7],
            [8, 9],
        ],
    )

    sfs = sfs.fit(X, y)
    assert sfs.k_feature_idx_ == (5, 10, 11, 12)


def test_knn_wo_cv_with_fixed_features_and_feature_groups_case1():
    # features (0, 1) gives different score?
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs = SFS(
        knn,
        k_features=(1, 2),
        scoring="accuracy",
        cv=0,
        fixed_features=[0, 1],
        feature_groups=[[0, 1], [2], [3]],
    )
    sfs.fit(X, y)
    # expect is based on what provided in `test_knn_wo_cv`
    expect = {
        1: {
            "feature_idx": (0, 1),
            "feature_names": ("0", "1"),
            "avg_score": 0.8333333333333334,
            "cv_scores": np.array([0.8333333333333334]),
        },
        2: {
            "feature_idx": (0, 1, 3),
            "feature_names": ("0", "1", "3"),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
    }
    dict_compare_utility(d_actual=expect, d_desired=sfs.subsets_)


def test_knn_wo_cv_with_fixed_features_and_feature_groups_case2():
    # similar to case1, but `fixed_features` is now consisting of two groups
    # [0,1] and [3]
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    efs1 = SFS(
        knn,
        k_features=2,
        scoring="accuracy",
        cv=0,
        fixed_features=[0, 1, 3],
        feature_groups=[[0, 1], [2], [3]],
    )
    efs1 = efs1.fit(X, y)
    # expect is based on what provided in `test_knn_wo_cv`
    expect = {
        2: {
            "feature_idx": (0, 1, 3),
            "feature_names": ("0", "1", "3"),
            "avg_score": 0.96666666666666667,
            "cv_scores": np.array([0.96666667]),
        },
    }
    dict_compare_utility(d_actual=expect, d_desired=efs1.subsets_)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from mlxtend.evaluate import BootstrapOutOfBag
from mlxtend.utils import assert_raises


def test_defaults():
    oob = BootstrapOutOfBag()
    results = list(oob.split(np.array([1, 2, 3, 4, 5])))
    assert len(results) == 200


def test_splits():
    oob = BootstrapOutOfBag(n_splits=3, random_seed=123)
    results = list(oob.split(np.array([1, 2, 3, 4, 5])))
    assert len(results) == 3
    assert np.array_equal(results[0][0], np.array([2, 4, 2, 1, 3]))
    assert np.array_equal(results[0][1], np.array([0]))
    assert np.array_equal(results[-1][0], np.array([1, 1, 0, 0, 1]))
    assert np.array_equal(results[-1][1], np.array([2, 3, 4]))


def test_invalid_splits():
    assert_raises(
        ValueError, "Number of splits must be greater than 1.", BootstrapOutOfBag, 0
    )


def test_get_n_splits():
    oob = BootstrapOutOfBag(n_splits=3, random_seed=123)
    assert oob.n_splits == 3



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import sys

if sys.version_info >= (3, 0):
    from .names import generalize_names
    from .names import generalize_names_duplcheck

from .tokenizer import tokenizer_emoticons, tokenizer_words_and_emoticons

__all__ = [
    "generalize_names",
    "generalize_names_duplcheck",
    "tokenizer_words_and_emoticons",
    "tokenizer_emoticons",
]



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import sys

import numpy as np
from sklearn.base import clone

from mlxtend.classifier import LogisticRegression
from mlxtend.data import iris_data
from mlxtend.utils import assert_raises

X, y = iris_data()
X = X[:, [0, 3]]  # sepal length and petal width
X = X[0:100]  # class 0 and class 1
y = y[0:100]  # class 0 and class 1

# standardize
X[:, 0] = (X[:, 0] - X[:, 0].mean()) / X[:, 0].std()
X[:, 1] = (X[:, 1] - X[:, 1].mean()) / X[:, 1].std()


def test_invalid_labels_1():
    y1 = np.where(y == 0, 2, 1)
    lr = LogisticRegression(epochs=15, eta=0.01, random_seed=1)

    if sys.version_info >= (3, 0):
        objtype = "{(0, 1)}"
    else:
        objtype = "set([(0, 1)])"

    expect = "Labels not in %s.\nFound (1, 2)" % objtype

    assert_raises(AttributeError, expect, lr.fit, X, y1, {(0, 1)})


def test_invalid_labels_2():
    y1 = np.where(y == 0, -1, 1)
    lr = LogisticRegression(epochs=15, eta=0.01, random_seed=1)
    assert_raises(
        AttributeError,
        "y array must not contain negative labels.\nFound [-1  1]",
        lr.fit,
        X,
        y1,
        {(-1, 1)},
    )


def test_logistic_regression_gd():
    w = np.array([[1.2], [4.4]])
    b = np.array([0.52])
    lr = LogisticRegression(epochs=100, eta=0.01, minibatches=1, random_seed=1)

    lr.fit(X, y)
    np.testing.assert_almost_equal(lr.w_, w, 2)
    np.testing.assert_almost_equal(lr.b_, b, 2)
    y_pred = lr.predict(X)
    acc = np.sum(y == y_pred, axis=0) / float(X.shape[0])
    assert acc == 1.0, "Acc: %s" % acc


def test_print_progress_1():
    lr = LogisticRegression(
        epochs=100, eta=0.01, minibatches=1, print_progress=1, random_seed=1
    )
    lr.fit(X, y)


def test_print_progress_2():
    lr = LogisticRegression(
        epochs=100, eta=0.01, minibatches=1, print_progress=2, random_seed=1
    )
    lr.fit(X, y)


def test_print_progress_3():
    lr = LogisticRegression(
        epochs=100, eta=0.01, minibatches=1, print_progress=3, random_seed=1
    )
    lr.fit(X, y)


def test_score_function():
    lr = LogisticRegression(epochs=100, eta=0.01, minibatches=1, random_seed=1)

    lr.fit(X, y)
    acc = lr.score(X, y)
    assert acc == 1.0, "Acc: %s" % acc


def test_refit_weights():
    w = np.array([[1.2], [4.4]])
    b = np.array([0.52])
    lr = LogisticRegression(epochs=50, eta=0.01, minibatches=1, random_seed=1)

    lr.fit(X, y)
    w1 = lr.w_[0][0]
    w2 = lr.w_[0][0]
    lr.fit(X, y, init_params=False)

    assert w1 != lr.w_[0][0]
    assert w2 != lr.w_[1][0]
    np.testing.assert_almost_equal(lr.w_, w, 2)
    np.testing.assert_almost_equal(lr.b_, b, 2)


def test_predict_proba():
    lr = LogisticRegression(epochs=100, eta=0.01, minibatches=1, random_seed=1)

    lr.fit(X, y)
    idx = [0, 48, 99]  # sample labels: 0, 0, 1
    y_pred = lr.predict_proba(X[idx])
    expect = np.array([0.009, 0.012, 0.993])
    np.testing.assert_almost_equal(y_pred, expect, 3)


def test_logistic_regression_sgd():
    w = np.array([[1.18], [4.38]])
    lr = LogisticRegression(epochs=100, eta=0.01, minibatches=len(y), random_seed=1)

    lr.fit(X, y)  # 0, 1 class
    np.testing.assert_almost_equal(lr.w_, w, 2)
    y_pred = lr.predict(X)
    acc = np.sum(y == y_pred, axis=0) / float(X.shape[0])
    assert acc == 1.0, "Acc: %s" % acc


def test_l2_regularization_gd():
    lr = LogisticRegression(
        eta=0.01, epochs=20, minibatches=1, l2_lambda=1.0, random_seed=1
    )
    lr.fit(X, y)
    y_pred = lr.predict(X)
    expect_weights = np.array([[1.061], [2.280]])

    np.testing.assert_almost_equal(lr.w_, expect_weights, 3)
    y_pred = lr.predict(X)
    acc = np.sum(y == y_pred, axis=0) / float(X.shape[0])
    assert acc == 1.0, "Acc: %s" % acc


def test_l2_regularization_sgd():
    lr = LogisticRegression(
        eta=0.01, epochs=100, minibatches=len(y), l2_lambda=1.0, random_seed=1
    )
    lr.fit(X, y)
    y_pred = lr.predict(X)
    expect_weights = np.array([[0.24], [0.35]])

    np.testing.assert_almost_equal(lr.w_, expect_weights, 2)
    y_pred = lr.predict(X)
    acc = np.sum(y == y_pred, axis=0) / float(X.shape[0])
    assert acc == 0.97, "Acc: %s" % acc


def test_ary_persistency_in_shuffling():
    orig = X.copy()
    lr = LogisticRegression(
        eta=0.01, epochs=100, minibatches=len(y), l2_lambda=1.0, random_seed=1
    )
    lr.fit(X, y)
    np.testing.assert_almost_equal(orig, X, 6)


def test_clone():
    log = LogisticRegression()
    clone(log)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from sklearn import datasets
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeClassifier

from mlxtend.plotting import plot_learning_curves


def test_training_size():
    iris = datasets.load_iris()
    X = iris.data
    y = iris.target
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.4, random_state=2
    )

    clf = DecisionTreeClassifier(max_depth=1, random_state=1)
    training_errors, test_errors = plot_learning_curves(
        X_train, y_train, X_test, y_test, clf, suppress_plot=True
    )

    desired1 = [0.22, 0.22, 0.22, 0.31, 0.31, 0.3, 0.33, 0.32, 0.33, 0.32]
    desired2 = [0.45, 0.45, 0.35, 0.35, 0.45, 0.43, 0.35, 0.35, 0.35, 0.35]

    np.testing.assert_almost_equal(training_errors, desired1, decimal=2)
    np.testing.assert_almost_equal(test_errors, desired2, decimal=2)


def test_scikit_metrics():
    iris = datasets.load_iris()
    X = iris.data
    y = iris.target
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.4, random_state=2
    )

    clf = DecisionTreeClassifier(max_depth=1, random_state=1)
    training_acc, test_acc = plot_learning_curves(
        X_train, y_train, X_test, y_test, clf, scoring="accuracy", suppress_plot=True
    )

    desired1 = np.array([0.22, 0.22, 0.22, 0.31, 0.31, 0.3, 0.33, 0.32, 0.33, 0.32])
    desired2 = np.array([0.45, 0.45, 0.35, 0.35, 0.45, 0.43, 0.35, 0.35, 0.35, 0.35])
    np.testing.assert_almost_equal(training_acc, 1 - desired1, decimal=2)
    np.testing.assert_almost_equal(test_acc, 1 - desired2, decimal=2)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import pytest

from mlxtend._base import _BaseModel, _Classifier


class BlankClassifier(_BaseModel, _Classifier):
    def __init__(self, print_progress=0, random_seed=1):
        self.print_progress = print_progress
        self.random_seed = random_seed

    def _fit(self, X, y, init_params=True):
        pass

    def _predict(self, X):
        pass


def test_init():
    cl = BlankClassifier(print_progress=0, random_seed=1)
    assert hasattr(cl, "print_progress")
    assert hasattr(cl, "random_seed")


def test_check_labels_ok_1():
    y = np.array([1, 1, 0])
    cl = BlankClassifier(print_progress=0, random_seed=1)
    cl._check_target_array(y=y, allowed={(0, 1)})


def test_check_labels_ok_2():
    y = np.array([1, 2, 2])
    cl = BlankClassifier(print_progress=0, random_seed=1)
    cl._check_target_array(y=y, allowed={(1, 2), (0, 1)})


def test_check_labels_not_ok_1():
    y = np.array([1, 3, 2])
    cl = BlankClassifier(print_progress=0, random_seed=1)
    with pytest.raises(AttributeError) as excinfo:
        cl._check_target_array(y, {(0, 1), (1, 2)})
        assert excinfo.value.message == (
            "Labels not in {(1, 2), (0, 1)}" ".\nFound (1, 2, 3)"
        )


def test_check_labels_integer_notok():
    y = np.array([1.0, 2.0], dtype=np.float_)
    cl = BlankClassifier(print_progress=0, random_seed=1)
    with pytest.raises(AttributeError) as excinfo:
        cl._check_target_array(y)
        assert excinfo.value.message == (
            "y must be an integer" " array.\nFound float64"
        )


def test_check_labels_positive_notok():
    y = np.array([1, 1, -1])
    cl = BlankClassifier(print_progress=0, random_seed=1)
    with pytest.raises(AttributeError) as excinfo:
        cl._check_target_array(y)
        assert excinfo.value.message == (
            "y array must not " "contain negative " "labels.\nFound [-1  1]"
        )


def test_predict_fail():
    X = np.array([[1], [2], [3]])
    est = BlankClassifier(print_progress=0, random_seed=1)
    est._is_fitted = False
    with pytest.raises(AttributeError) as excinfo:
        est.predict(X)
        assert excinfo.value.message == ("Model is not " "fitted, yet.")


def test_predict_pass():
    X = np.array([[1], [2], [3]])
    y = np.array([1, 2, 3])
    est = BlankClassifier(print_progress=0, random_seed=1)
    est.fit(X, y)
    est.predict(X)

    with pytest.raises(TypeError) as excinfo:
        est.fit(X)
        assert excinfo.value.message == (
            "fit() missing 1 " "required positional " "argument: 'y'"
        )


def test_fit_2():
    X = np.array([[1], [2], [3]])
    y = np.array([1, 2, 3])
    est = BlankClassifier(print_progress=0, random_seed=1)
    est.fit(X=X, y=y)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# A class to apply column-based mean centering to a dataset.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np


class MeanCenterer(object):

    """Column centering of vectors and matrices.

    Attributes
    -----------
    col_means : numpy.ndarray [n_columns]
        NumPy array storing the mean values for centering after fitting
        the MeanCenterer object.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/preprocessing/MeanCenterer/

    """

    def __init__(self):
        pass

    def transform(self, X):
        """Centers a NumPy array.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Array of data vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        X_tr : {array-like, sparse matrix}, shape = [n_samples, n_features]
            A copy of the input array with the columns centered.

        """
        if not hasattr(self, "col_means"):
            raise AttributeError("MeanCenterer has not been fitted, yet.")
        X_tr = np.copy(self._get_array(X))
        X_tr = np.apply_along_axis(
            func1d=lambda x: x - self.col_means, axis=1, arr=X_tr
        )
        return X_tr

    def fit(self, X):
        """Gets the column means for mean centering.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Array of data vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self
        """
        self.col_means = self._get_array(X).mean(axis=0)
        return self

    def fit_transform(self, X):
        """Fits and transforms an arry.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Array of data vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        X_tr : {array-like, sparse matrix}, shape = [n_samples, n_features]
            A copy of the input array with the columns centered.
        """
        self.fit(X)
        return self.transform(X)

    def _get_array(self, X):
        if isinstance(X, list):
            X_fl = np.asarray(X, dtype="float")[:, None]
        else:
            X_fl = X.astype("float")
        return X_fl



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Base MultiClass (MultiClass Parent Class)
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np


class _MultiClass(object):
    def __init__(self):
        pass

    def _one_hot(self, y, n_labels, dtype):
        """Returns a matrix where each sample in y is represented
           as a row, and each column represents the class label in
           the one-hot encoding scheme.

        Example:

            y = np.array([0, 1, 2, 3, 4, 2])
            mc = _BaseMultiClass()
            mc._one_hot(y=y, n_labels=5, dtype='float')

            np.array([[1., 0., 0., 0., 0.],
                      [0., 1., 0., 0., 0.],
                      [0., 0., 1., 0., 0.],
                      [0., 0., 0., 1., 0.],
                      [0., 0., 0., 0., 1.],
                      [0., 0., 1., 0., 0.]])

        """
        mat = np.zeros((len(y), n_labels))
        for i, val in enumerate(y):
            mat[i, val] = 1
        return mat.astype(dtype)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# A function for fetching the open-source MNIST dataset.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import struct

import numpy as np


def loadlocal_mnist(images_path, labels_path):
    """Read MNIST from ubyte files.

    Parameters
    ----------
    images_path : str
        path to the test or train MNIST ubyte file
    labels_path : str
        path to the test or train MNIST class labels file

    Returns
    --------
    images : [n_samples, n_pixels] numpy.array
        Pixel values of the images.
    labels : [n_samples] numpy array
        Target class labels

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/data/loadlocal_mnist/

    """
    with open(labels_path, "rb") as lbpath:
        magic, n = struct.unpack(">II", lbpath.read(8))
        labels = np.fromfile(lbpath, dtype=np.uint8)
    with open(images_path, "rb") as imgpath:
        magic, num, rows, cols = struct.unpack(">IIII", imgpath.read(16))
        images = np.fromfile(imgpath, dtype=np.uint8).reshape(len(labels), 784)

    return images, labels



# PyPrind
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# Contributors: https://github.com/rasbt/pyprind/graphs/contributors
# License: BSD 3 clause
# Code Repository: https://github.com/rasbt/pyprind
# PyPI: https://pypi.python.org/pypi/PyPrind

from .progbar import ProgBar
from .progpercent import ProgPercent


def generator_factory(mother_class):
    def generator_progress(iteritem, iterations=None, *args, **kw):
        if iterations is None:
            iterations = len(iteritem)
        assert iterations
        mbar = mother_class(iterations, *args, **kw)
        for item in iteritem:
            yield item
            mbar.update()

    return generator_progress


prog_percent = generator_factory(ProgPercent)
prog_bar = generator_factory(ProgBar)



# Sebastian Raschka 2014-2016
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from numpy.testing import assert_array_equal, assert_raises

from mlxtend.evaluate import lift_score
from mlxtend.evaluate.lift_score import support


def test_multiclass():
    y_targ = [1, 1, 1, 0, 0, 2, 0, 3, 4]
    y_pred = [1, 0, 1, 0, 0, 2, 1, 3, 0]
    x = 2
    y = lift_score(y_targ, y_pred, binary=True, positive_label=1)
    assert_array_equal(x, y)


def test_multiclass_positive_label_0():
    y_targ = [1, 1, 1, 0, 0, 2, 0, 3, 4]
    y_pred = [1, 0, 1, 0, 0, 2, 1, 3, 0]
    x = 1.5
    y = lift_score(y_targ, y_pred, binary=True, positive_label=0)
    assert_array_equal(x, y)


def test_multiclass_with_false_binary():
    y_targ = [1, 1, 1, 0, 0, 2, 0, 3]
    y_pred = [1, 0, 1, 0, 0, 2, 1, 3]
    assert_raises(AttributeError, lift_score, y_targ, y_pred, False, 1)


def test_binary_with_numpy():
    y_targ = np.array([1, 1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 0])
    y_pred = np.array([1, 0, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0])
    x = 1.25
    y = lift_score(y_targ, y_pred, binary=False, positive_label=1)
    assert_array_equal(x, y)


def test_binary():
    y_targ = [1, 1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 0]
    y_pred = [1, 0, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0]
    x = 1.25
    y = lift_score(y_targ, y_pred, binary=False, positive_label=1)
    assert_array_equal(x, y)


def test_multidimension():
    y_targ = [[1, 1, 1, 0, 0, 1], [0, 1, 0, 0, 0, 1]]
    y_pred = [[1, 0, 1, 0, 0, 1]]
    x = 1
    y = lift_score(y_targ, y_pred, binary=False, positive_label=1)
    assert_array_equal(x, y)


def test_support_with_two_parameter():
    y_targ = np.array([1, 1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 0])
    y_pred = np.array([1, 0, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0])
    x = 0.4
    y = support(y_targ, y_pred)
    assert_array_equal(x, y)


def test_support_with_one_parameter():
    y_targ = np.array([1, 1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 0])
    x = 0.533333333333333333
    y = support(y_targ)
    assert_array_equal(x, y)



import numpy as np
from scipy import sparse
from sklearn.base import ClassifierMixin

from ..externals.estimator_checks import check_is_fitted


class _BaseStackingClassifier(ClassifierMixin):
    """Base class of stacking classifiers"""

    def _do_predict(self, X, predict_fn):
        meta_features = self.predict_meta_features(X)

        if not self.use_features_in_secondary:
            return predict_fn(meta_features)
        elif sparse.issparse(X):
            return predict_fn(sparse.hstack((X, meta_features)))
        else:
            return predict_fn(np.hstack((X, meta_features)))

    def predict(self, X):
        """Predict target values for X.

        Parameters
        ----------
        X : numpy array, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        ----------
        labels : array-like, shape = [n_samples]
            Predicted class labels.

        """
        check_is_fitted(self, ["clfs_", "meta_clf_"])

        return self._do_predict(X, self.meta_clf_.predict)

    def predict_proba(self, X):
        """ Predict class probabilities for X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        ----------
        proba : array-like, shape = [n_samples, n_classes] or a list of \
                n_outputs of such arrays if n_outputs > 1.
            Probability for each class per sample.

        """
        check_is_fitted(self, ["clfs_", "meta_clf_"])

        return self._do_predict(X, self.meta_clf_.predict_proba)

    def decision_function(self, X):
        """ Predict class confidence scores for X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        ----------
        scores : shape=(n_samples,) if n_classes == 2 else \
            (n_samples, n_classes).
            Confidence scores per (sample, class) combination. In the binary
            case, confidence score for self.classes_[1] where >0 means this
            class would be predicted.

        """
        check_is_fitted(self, ["clfs_", "meta_clf_"])

        return self._do_predict(X, self.meta_clf_.decision_function)



# Stacking regressor

# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# An ensemble-learning meta-regressor for stacking regression
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import scipy.sparse as sparse
from sklearn.base import RegressorMixin, TransformerMixin, clone
from sklearn.utils import check_X_y

from ..externals.estimator_checks import check_is_fitted
from ..externals.name_estimators import _name_estimators
from ..utils.base_compostion import _BaseXComposition


class StackingRegressor(_BaseXComposition, RegressorMixin, TransformerMixin):

    """A Stacking regressor for scikit-learn estimators for regression.

    Parameters
    ----------
    regressors : array-like, shape = [n_regressors]
        A list of regressors.
        Invoking the `fit` method on the `StackingRegressor` will fit clones
        of those original regressors that will
        be stored in the class attribute
        `self.regr_`.

    meta_regressor : object
        The meta-regressor to be fitted on the ensemble of
        regressors

    verbose : int, optional (default=0)
        Controls the verbosity of the building process.
        - `verbose=0` (default): Prints nothing
        - `verbose=1`: Prints the number & name of the regressor being fitted
        - `verbose=2`: Prints info about the parameters of the
                       regressor being fitted
        - `verbose>2`: Changes `verbose` param of the underlying regressor to
           self.verbose - 2

    use_features_in_secondary : bool (default: False)
        If True, the meta-regressor will be trained both on
        the predictions of the original regressors and the
        original dataset.
        If False, the meta-regressor will be trained only on
        the predictions of the original regressors.

    store_train_meta_features : bool (default: False)
        If True, the meta-features computed from the training data
        used for fitting the
        meta-regressor stored in the `self.train_meta_features_` array,
        which can be
        accessed after calling `fit`.


    Attributes
    ----------
    regr_ : list, shape=[n_regressors]
        Fitted regressors (clones of the original regressors)

    meta_regr_ : estimator
        Fitted meta-regressor (clone of the original meta-estimator)

    coef_ : array-like, shape = [n_features]
        Model coefficients of the fitted meta-estimator

    intercept_ : float
        Intercept of the fitted meta-estimator

    train_meta_features : numpy array,
        shape = [n_samples, len(self.regressors)]
        meta-features for training data, where n_samples is the
        number of samples
        in training data and len(self.regressors) is the number of regressors.

    refit : bool (default: True)
        Clones the regressors for stacking regression if True (default)
        or else uses the original ones, which will be refitted on the dataset
        upon calling the `fit` method. Setting refit=False is
        recommended if you are working with estimators that are supporting
        the scikit-learn fit/predict API interface but are not compatible
        to scikit-learn's `clone` function.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/regressor/StackingRegressor/

    """

    def __init__(
        self,
        regressors,
        meta_regressor,
        verbose=0,
        use_features_in_secondary=False,
        store_train_meta_features=False,
        refit=True,
        multi_output=False,
    ):
        self.regressors = regressors
        self.meta_regressor = meta_regressor
        self.verbose = verbose
        self.use_features_in_secondary = use_features_in_secondary
        self.store_train_meta_features = store_train_meta_features
        self.refit = refit
        self.multi_output = multi_output

    @property
    def named_regressors(self):
        return _name_estimators(self.regressors)

    def fit(self, X, y, sample_weight=None):
        """Learn weight coefficients from training data for each regressor.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        y : numpy array, shape = [n_samples] or [n_samples, n_targets]
             Target values. Multiple targets are supported only if
             self.multi_output is True.

        sample_weight : array-like, shape = [n_samples], optional
            Sample weights passed as sample_weights to each regressor
            in the regressors list as well as the meta_regressor.
            Raises error if some regressor does not support
            sample_weight in the fit() method.

        Returns
        -------
        self : object

        """
        X, y = check_X_y(
            X,
            y,
            accept_sparse=["csc", "csr"],
            dtype=None,
            multi_output=self.multi_output,
        )

        if self.refit:
            self.regr_ = clone(self.regressors)
            self.meta_regr_ = clone(self.meta_regressor)
        else:
            self.regr_ = self.regressors
            self.meta_regr_ = self.meta_regressor

        if self.verbose > 0:
            print("Fitting %d regressors..." % (len(self.regressors)))

        for regr in self.regr_:
            if self.verbose > 0:
                i = self.regr_.index(regr) + 1
                print(
                    "Fitting regressor%d: %s (%d/%d)"
                    % (i, _name_estimators((regr,))[0][0], i, len(self.regr_))
                )

            if self.verbose > 2:
                if hasattr(regr, "verbose"):
                    regr.set_params(verbose=self.verbose - 2)

            if self.verbose > 1:
                print(_name_estimators((regr,))[0][1])

            if sample_weight is None:
                regr.fit(X, y)
            else:
                regr.fit(X, y, sample_weight=sample_weight)

        meta_features = self.predict_meta_features(X)

        if not self.use_features_in_secondary:
            # meta model uses the prediction outcomes only
            pass
        elif sparse.issparse(X):
            meta_features = sparse.hstack((X, meta_features))
        else:
            meta_features = np.hstack((X, meta_features))

        if sample_weight is None:
            self.meta_regr_.fit(meta_features, y)
        else:
            self.meta_regr_.fit(meta_features, y, sample_weight=sample_weight)

        # save meta-features for training data
        if self.store_train_meta_features:
            self.train_meta_features_ = meta_features
        return self

    @property
    def coef_(self):
        return self.meta_regr_.coef_

    @property
    def intercept_(self):
        return self.meta_regr_.intercept_

    def get_params(self, deep=True):
        """Return estimator parameter names for GridSearch support."""
        return self._get_params("named_regressors", deep=deep)

    def set_params(self, **params):
        """Set the parameters of this estimator.

        Valid parameter keys can be listed with ``get_params()``.

        Returns
        -------
        self
        """
        self._set_params("regressors", "named_regressors", **params)
        return self

    def predict_meta_features(self, X):
        """Get meta-features of test-data.

        Parameters
        ----------
        X : numpy array, shape = [n_samples, n_features]
            Test vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        meta-features : numpy array, shape = [n_samples, len(self.regressors)]
            meta-features for test data, where n_samples is the number of
            samples in test data and len(self.regressors) is the number
            of regressors. If self.multi_output is True, then the number of
            columns is len(self.regressors) * n_targets

        """
        check_is_fitted(self, "regr_")
        return np.column_stack([r.predict(X) for r in self.regr_])

    def predict(self, X):
        """Predict target values for X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        ----------
        y_target : array-like, shape = [n_samples] or [n_samples, n_targets]
            Predicted target values.
        """
        check_is_fitted(self, "regr_")
        meta_features = self.predict_meta_features(X)

        if not self.use_features_in_secondary:
            return self.meta_regr_.predict(meta_features)
        elif sparse.issparse(X):
            return self.meta_regr_.predict(sparse.hstack((X, meta_features)))
        else:
            return self.meta_regr_.predict(np.hstack((X, meta_features)))
